from __future__ import division
import math
import pprint
import random

class DependencyResolver:
    def solve(self, arg):
        d = dict((k, set(arg[k])) for k in arg)
        r = []
        while d:
            # values not in keys (items without dep)
            t = set(i for v in d.values() for i in v) - set(d.keys())
            # and keys without value (items without dep)
            t.update(k for k, v in d.items() if not v)
            # can be done right away
            if t:
                r.append(t)
            else:
                # d contains the the data which are useful to resolve the issue 
                raise RuntimeError("Cycle detect {}".format(pprint.pformat(d)))
            # and cleaned up
            d = dict(((k, set(v) - t) for k, v in d.items() if v))
        return r

class ParserError(Exception):
    pass

class InvalidExpressionError(ParserError):
    pass

class ParsingError(ParserError):
    pass

class EpressionParityError(ParserError):
    pass

class UndefinedVariableError(ParserError):
    pass

class NotAFunctionError(ParserError):
    pass


class Token(object):

    TNUMBER = 0
    TOP1 = 1
    TOP2 = 2
    TVAR = 3
    TFUNCALL = 4

    def __init__(self, type_, index_, prio_, number_):
        self.type_ = type_
        self.index_ = index_ or 0
        self.prio_ = prio_ or 0
        self.number_ = number_ if number_ != None else 0

    def toString(self):
        if self.type_ == Token.TNUMBER:
            return self.number_
        if self.type_ == Token.TOP1 or self.type_ == Token.TOP2 or self.type_ == Token.TVAR:
            return self.index_
        elif self.type_ == Token.TFUNCALL:
            return 'CALL'
        else:
            return 'Invalid Token'


class Expression(object):
    def __init__(self, tokens, ops1, ops2, functions):
        self.tokens = tokens
        self.ops1 = ops1
        self.ops2 = ops2
        self.functions = functions

    def simplify(self, values):
        values = values or {}
        nstack = []
        newexpression = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == Token.TNUMBER:
                nstack.append(item)
            elif type_ == Token.TVAR and item.index_ in values:
                item = Token(Token.TNUMBER, 0, 0, values[item.index_])
                nstack.append(item)
            elif type_ == Token.TOP2 and len(nstack) > 1:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = self.ops2[item.index_]
                item = Token(Token.TNUMBER, 0, 0, f(n1.number_, n2.number_))
                nstack.append(item)
            elif type_ == Token.TOP1 and nstack:
                n1 = nstack.pop()
                f = self.ops1[item.index_]
                item = Token(Token.TNUMBER, 0, 0, f(n1.number_))
                nstack.append(item)
            else:
                while len(nstack) > 0:
                    newexpression.append(nstack.pop(0))
                newexpression.append(item)
        while nstack:
            newexpression.append(nstack.pop(0))

        return Expression(newexpression, self.ops1, self.ops2, self.functions)

    def substitute(self, variable, expr):
        if not isinstance(expr, Expression):
            expr = Parser().parse(str(expr))
        newexpression = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == Token.TVAR and item.index_ == variable:
                for j in range(0, len(expr.tokens)):
                    expritem = expr.tokens[j]
                    replitem = Token(
                        expritem.type_,
                        expritem.index_,
                        expritem.prio_,
                        expritem.number_,
                    )
                    newexpression.append(replitem)
            else:
                newexpression.append(item)

        ret = Expression(newexpression, self.ops1, self.ops2, self.functions)
        return ret

    def evaluate(self, values):
        values = values or {}
        nstack = []
        L = len(self.tokens)
        for item in self.tokens:
            type_ = item.type_
            if type_ == Token.TNUMBER:
                nstack.append(item.number_)
            elif type_ == Token.TOP2:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = self.ops2[item.index_]
                nstack.append(f(n1, n2))
            elif type_ == Token.TVAR:
                if item.index_ in values:
                    nstack.append(values[item.index_])
                elif item.index_ in self.functions:
                    nstack.append(self.functions[item.index_])
                else:
                    raise UndefinedVariableError('undefined variable: ' + item.index_)
            elif type_ == Token.TOP1:
                n1 = nstack.pop()
                f = self.ops1[item.index_]
                nstack.append(f(n1))
            elif type_ == Token.TFUNCALL:
                n1 = nstack.pop()
                f = nstack.pop()
                if callable(f):
                    if type(n1) is list:
                        nstack.append(f(*n1))
                    else:
                        nstack.append(f(n1))
                else:
                    raise NotAFunctionError(f + ' is not a function')
            else:
                raise InvalidExpressionError('invalid Expression')
        if len(nstack) > 1:
            raise EpressionParityError('invalid Expression (parity)')
        return nstack[0]

    def toString(self, toJS=False):
        nstack = []
        L = len(self.tokens)
        for i in range(0, L):
            item = self.tokens[i]
            type_ = item.type_
            if type_ == Token.TNUMBER:
                if type(item.number_) == str:
                    nstack.append("'" + item.number_ + "'")
                else:
                    nstack.append(item.number_)
            elif type_ == Token.TOP2:
                n2 = nstack.pop()
                n1 = nstack.pop()
                f = item.index_
                if toJS and f == '^':
                    nstack.append('math.pow(' + n1 + ',' + n2 + ')')
                else:
                    frm = '({n1}{f}{n2})'
                    if f == ',':
                        frm = '{n1}{f}{n2}'

                    nstack.append(frm.format(
                        n1=n1,
                        n2=n2,
                        f=f,
                    ))


            elif type_ == Token.TVAR:
                nstack.append(item.index_)
            elif type_ == Token.TOP1:
                n1 = nstack.pop()
                f = item.index_
                if f == '-':
                    nstack.append('(' + f + n1 + ')')
                else:
                    nstack.append(f + '(' + n1 + ')')
            elif type_ == Token.TFUNCALL:
                n1 = nstack.pop()
                f = nstack.pop()
                nstack.append(f + '(' + n1 + ')')
            else:
                raise InvalidExpressionError('invalid Expression')
        if len(nstack) > 1:
            raise EpressionParityError('invalid Expression (parity)')
        return nstack[0]

    def __str__(self):
        return self.toString()

    def symbols(self):
        vars = []
        for i in range(0, len(self.tokens)):
            item = self.tokens[i]
            if item.type_ == Token.TVAR and not item.index_ in vars:
                vars.append(item.index_)
        return vars

    def variables(self):
        return [
            sym for sym in self.symbols()
            if sym not in self.functions]


class Parser:
    class Expression(Expression):
        pass

    PRIMARY = 1
    OPERATOR = 2
    FUNCTION = 4
    LPAREN = 8
    RPAREN = 16
    COMMA = 32
    SIGN = 64
    CALL = 128
    NULLARY_CALL = 256

    def add(self, a, b):
        return a + b

    def sub(self, a, b):
        return a - b

    def mul(self, a, b):
        return a * b

    def div(self, a, b):
        return a / b

    def mod(self, a, b):
        return a % b

    def concat(self, a, b, *args):
        result = u'{0}{1}'.format(a, b)
        for arg in args:
            result = u'{0}{1}'.format(result, arg)
        return result

    def equal(self, a, b):
        return a == b

    def notEqual(self, a, b):
        return a != b

    def greaterThan(self, a, b):
        return a > b

    def lessThan(self, a, b):
        return a < b

    def greaterThanEqual(self, a, b):
        return a >= b

    def lessThanEqual(self, a, b):
        return a <= b

    def andOperator(self, a, b):
        return (a and b)

    def orOperator(self, a, b):
        return (a or b)

    def neg(self, a):
        return -a

    def random(self, a):
        return math.random() * (a or 1)

    def fac(self, a):  # a!
        return math.factorial(a)

    def pyt(self, a, b):
        return math.sqrt(a * a + b * b)

    def ifFunction(self, a, b, c):
        return b if a else c

    def append(self, a, b):
        if type(a) != list:
            return [a, b]
        a.append(b)
        return a

    def __init__(self):
        self.success = False
        self.errormsg = ''
        self.expression = ''

        self.pos = 0

        self.tokennumber = 0
        self.tokenprio = 0
        self.tokenindex = 0
        self.tmpprio = 0

        self.ops1 = {
            'sin': math.sin,
            'cos': math.cos,
            'tan': math.tan,
            'asin': math.asin,
            'acos': math.acos,
            'atan': math.atan,
            'sqrt': math.sqrt,
            'log': math.log,
            'abs': abs,
            'ceil': math.ceil,
            'floor': math.floor,
            'round': round,
            '-': self.neg,
            'exp': math.exp,
        }

        self.ops2 = {
            '+': self.add,
            '-': self.sub,
            '*': self.mul,
            '/': self.div,
            '%': self.mod,
            '^': math.pow,
            ',': self.append,
            '||': self.concat,
            "==": self.equal,
            "!=": self.notEqual,
            ">": self.greaterThan,
            "<": self.lessThan,
            ">=": self.greaterThanEqual,
            "<=": self.lessThanEqual,
            "and": self.andOperator,
            "or": self.orOperator
        }

        self.functions = {
            'random': random,
            'fac': self.fac,
            'min': min,
            'max': max,
            'pyt': self.pyt,
            'pow': math.pow,
            'atan2': math.atan2,
            'concat': self.concat,
            'if': self.ifFunction
        }

        self.consts = {
            'E': math.e,
            'PI': math.pi,
        }

        self.values = {
            'sin': math.sin,
            'cos': math.cos,
            'tan': math.tan,
            'asin': math.asin,
            'acos': math.acos,
            'atan': math.atan,
            'sqrt': math.sqrt,
            'log': math.log,
            'abs': abs,
            'ceil': math.ceil,
            'floor': math.floor,
            'round': round,
            'random': self.random,
            'fac': self.fac,
            'exp': math.exp,
            'min': min,
            'max': max,
            'pyt': self.pyt,
            'pow': math.pow,
            'atan2': math.atan2,
            'E': math.e,
            'PI': math.pi
        }

    def parse(self, expr):
        self.errormsg = ''
        self.success = True
        operstack = []
        tokenstack = []
        self.tmpprio = 0
        expected = self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
        noperators = 0
        self.expression = expr
        self.pos = 0

        while self.pos < len(self.expression):
            if self.isOperator():
                if self.isSign() and expected & self.SIGN:
                    if self.isNegativeSign():
                        self.tokenprio = 5
                        self.tokenindex = '-'
                        noperators += 1
                        self.addfunc(tokenstack, operstack, Token.TOP1)
                    expected = \
                        self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
                elif self.isComment():
                    pass
                else:
                    if expected and self.OPERATOR == 0:
                        self.error_parsing(self.pos, 'unexpected operator')
                    noperators += 2
                    self.addfunc(tokenstack, operstack, Token.TOP2)
                    expected = \
                        self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
            elif self.isNumber():
                if expected and self.PRIMARY == 0:
                    self.error_parsing(self.pos, 'unexpected number')
                token = Token(Token.TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(token)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.isString():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected string')
                token = Token(Token.TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(token)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.isLeftParenth():
                if (expected & self.LPAREN) == 0:
                    self.error_parsing(self.pos, 'unexpected \"(\"')
                if expected & self.CALL:
                    noperators += 2
                    self.tokenprio = -2
                    self.tokenindex = -1
                    self.addfunc(tokenstack, operstack, Token.TFUNCALL)
                expected = \
                    self.PRIMARY | self.LPAREN | self.FUNCTION | \
                    self.SIGN | self.NULLARY_CALL
            elif self.isRightParenth():
                if expected & self.NULLARY_CALL:
                    token = Token(Token.TNUMBER, 0, 0, [])
                    tokenstack.append(token)
                elif (expected & self.RPAREN) == 0:
                    self.error_parsing(self.pos, 'unexpected \")\"')
                expected = \
                    self.OPERATOR | self.RPAREN | self.COMMA | \
                    self.LPAREN | self.CALL
            elif self.isComma():
                if (expected & self.COMMA) == 0:
                    self.error_parsing(self.pos, 'unexpected \",\"')
                self.addfunc(tokenstack, operstack, Token.TOP2)
                noperators += 2
                expected = \
                    self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
            elif self.isConst():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected constant')
                consttoken = Token(Token.TNUMBER, 0, 0, self.tokennumber)
                tokenstack.append(consttoken)
                expected = self.OPERATOR | self.RPAREN | self.COMMA
            elif self.isOp2():
                if (expected & self.FUNCTION) == 0:
                    self.error_parsing(self.pos, 'unexpected function')
                self.addfunc(tokenstack, operstack, Token.TOP2)
                noperators += 2
                expected = self.LPAREN
            elif self.isOp1():
                if (expected & self.FUNCTION) == 0:
                    self.error_parsing(self.pos, 'unexpected function')
                self.addfunc(tokenstack, operstack, Token.TOP1)
                noperators += 1
                expected = self.LPAREN
            elif self.isVar():
                if (expected & self.PRIMARY) == 0:
                    self.error_parsing(self.pos, 'unexpected variable')
                vartoken = Token(Token.TVAR, self.tokenindex, 0, 0)
                tokenstack.append(vartoken)
                expected = \
                    self.OPERATOR | self.RPAREN | \
                    self.COMMA | self.LPAREN | self.CALL
            elif self.isWhite():
                pass
            else:
                if self.errormsg == '':
                    self.error_parsing(self.pos, 'unknown character')
                else:
                    self.error_parsing(self.pos, self.errormsg)
        if self.tmpprio < 0 or self.tmpprio >= 10:
            self.error_parsing(self.pos, 'unmatched \"()\"')
        while len(operstack) > 0:
            tmp = operstack.pop()
            tokenstack.append(tmp)
        if (noperators + 1) != len(tokenstack):
            self.error_parsing(self.pos, 'parity')

        return Expression(tokenstack, self.ops1, self.ops2, self.functions)

    def evaluate(self, expr, variables):
        return self.parse(expr).evaluate(variables)

    def error_parsing(self, column, msg):
        self.success = False
        self.errormsg = 'parse error [column ' + str(column) + ']: ' + msg
        raise ParsingError(self.errormsg)

    def addfunc(self, tokenstack, operstack, type_):
        operator = Token(
            type_,
            self.tokenindex,
            self.tokenprio + self.tmpprio,
            0,
        )
        while len(operstack) > 0:
            if operator.prio_ <= operstack[len(operstack) - 1].prio_:
                tokenstack.append(operstack.pop())
            else:
                break
        operstack.append(operator)

    def isNumber(self):
        r = False
        str = ''
        while self.pos < len(self.expression):
            code = self.expression[self.pos]
            if (code >= '0' and code <= '9') or code == '.':
                str += self.expression[self.pos]
                self.pos += 1
                self.tokennumber = float(str)
                r = True
            else:
                break
        return r

    def unescape(self, v, pos):
        buffer = []
        escaping = False

        for i in range(0, len(v)):
            c = v[i]

            if escaping:
                if c == "'":
                    buffer.append("'")
                    break
                elif c == '\\':
                    buffer.append('\\')
                    break
                elif c == '/':
                    buffer.append('/')
                    break
                elif c == 'b':
                    buffer.append('\b')
                    break
                elif c == 'f':
                    buffer.append('\f')
                    break
                elif c == 'n':
                    buffer.append('\n')
                    break
                elif c == 'r':
                    buffer.append('\r')
                    break
                elif c == 't':
                    buffer.append('\t')
                    break
                elif c == 'u':
                    # interpret the following 4 characters
                    # as the hex of the unicode code point
                    codePoint = int(v[i + 1: i + 5], 16)
                    buffer.append(unichr(codePoint))
                    i += 4
                    break
                else:
                    self.error_parsing(
                        pos + i,
                        'Illegal escape sequence: \'\\' + c + '\'',
                    )
                escaping = False
            else:
                if c == '\\':
                    escaping = True
                else:
                    buffer.append(c)

        return ''.join(buffer)

    def isString(self):
        r = False
        str = ''
        startpos = self.pos
        if self.pos < len(self.expression) and self.expression[self.pos] == "'":
            self.pos += 1
            while self.pos < len(self.expression):
                code = self.expression[self.pos]
                if code != '\'' or (str != '' and str[-1] == '\\'):
                    str += self.expression[self.pos]
                    self.pos += 1
                else:
                    self.pos += 1
                    self.tokennumber = self.unescape(str, startpos)
                    r = True
                    break
        return r

    def isConst(self):
        for i in self.consts:
            L = len(i)
            str = self.expression[self.pos:self.pos + L]
            if i == str:
                if len(self.expression) <= self.pos + L:
                    self.tokennumber = self.consts[i]
                    self.pos += L
                    return True
                if not self.expression[self.pos + L].isalnum() and self.expression[self.pos + L] != "_":
                    self.tokennumber = self.consts[i]
                    self.pos += L
                    return True
        return False

    def isOperator(self):
        ops = (
            ('+', 2, '+'),
            ('-', 2, '-'),
            ('*', 3, '*'),
            (u'\u2219', 3, '*'),  # bullet operator
            (u'\u2022', 3, '*'),  # black small circle
            ('/', 4, '/'),
            ('%', 4, '%'),
            ('^', 6, '^'),
            ('||', 1, '||'),
            ('==', 1, '=='),
            ('!=', 1, '!='),
            ('<=', 1, '<='),
            ('>=', 1, '>='),
            ('<', 1, '<'),
            ('>', 1, '>'),
            ('and', 0, 'and'),
            ('or', 0, 'or'),
        )
        for token, priority, index in ops:
            if self.expression.startswith(token, self.pos):
                self.tokenprio = priority
                self.tokenindex = index
                self.pos += len(token)
                return True
        return False

    def isSign(self):
        code = self.expression[self.pos - 1]
        return (code == '+') or (code == '-')

    def isNegativeSign(self):
        code = self.expression[self.pos - 1]
        return code == '-'

    def isLeftParenth(self):
        code = self.expression[self.pos]
        if code == '(':
            self.pos += 1
            self.tmpprio += 10
            return True
        return False

    def isRightParenth(self):
        code = self.expression[self.pos]
        if code == ')':
            self.pos += 1
            self.tmpprio -= 10
            return True
        return False

    def isComma(self):
        code = self.expression[self.pos]
        if code == ',':
            self.pos += 1
            self.tokenprio = -1
            self.tokenindex = ","
            return True
        return False

    def isWhite(self):
        code = self.expression[self.pos]
        if code.isspace():
            self.pos += 1
            return True
        return False

    def isOp1(self):
        str = ''
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.upper() == c.lower():
                if i == self.pos or (c != '_' and (c < '0' or c > '9')):
                    break
            str += c
        if len(str) > 0 and str in self.ops1:
            self.tokenindex = str
            self.tokenprio = 7
            self.pos += len(str)
            return True
        return False

    def isOp2(self):
        str = ''
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.upper() == c.lower():
                if i == self.pos or (c != '_' and (c < '0' or c > '9')):
                    break
            str += c
        if len(str) > 0 and (str in self.ops2):
            self.tokenindex = str
            self.tokenprio = 7
            self.pos += len(str)
            return True
        return False

    def isVar(self):
        str = ''
        inQuotes = False
        for i in range(self.pos, len(self.expression)):
            c = self.expression[i]
            if c.lower() == c.upper():
                if ((i == self.pos and c != '"') or (not (c in '_."') and (c < '0' or c > '9'))) and not inQuotes:
                    if not (c == '_' and i == self.pos):
                        break
            if c == '"':
                inQuotes = not inQuotes
            str += c
        if str:
            self.tokenindex = str
            self.tokenprio = 4
            self.pos += len(str)
            return True
        return False

    def isComment(self):
        code = self.expression[self.pos - 1]
        if code == '/' and self.expression[self.pos] == '*':
            # the comment must be closed and not empty. If it is not closed, next line will raise exception 
            self.pos = self.expression.index('*/', self.pos) + 2
            return True
        return False



import math
import pytest

from FormulaToolkit import Parser
from FormulaToolkit import InvalidExpressionError, ParsingError, EpressionParityError, UndefinedVariableError, NotAFunctionError

@pytest.fixture()
def parser():
    parser = Parser()
    yield parser.evaluate
    del parser

def test_parser_trigonometric(parser):
    assert 1 == parser('sin(1)^2+cos(1)^2', {})
    assert 1 == parser('sin(x)^2+cos(x)^2', {'x':1})

def test_parser_max_min(parser):
    assert 10 == parser('max(1,2,3,4,5,6,7,8,10)', {})
    assert 1  == parser('min(1,2,3,4,5,6,7,8,10)', {})

def test_parser_if(parser):
    assert 'blue' == parser("if(0,'red', 'blue')",{'red': '', 'blue': ''})
    assert 'red'  == parser("if(1,'red', 'blue')",{'red': '', 'blue': ''})

def test_parser_constants(parser):
    assert math.e == parser('E', {})
    assert math.pi == parser('PI', {})

def test_parser_concat(parser):
    assert 'pippo' == parser("concat('pip', 'po')", {})
    assert 'pippo' == parser("concat('p','i','p','p','o')", {})

def test_parser_pitagora(parser):
    assert 5 == parser('pyt(a,b)',{'a': 3,'b':4})

def test_parser_nested_simple_1(parser):
    assert 6 == parser('a+(b+c)',{'a':1, 'b':2, 'c':3, 'd':4, 'e':5, 'f':6, 'g':7, 'h':8})

def test_parser_nested_simple_1(parser):
    assert 2.25 == parser('a+(b+c)/d',{'a':1, 'b':2, 'c':3, 'd':4, 'e':5, 'f':6, 'g':7, 'h':8})

def test_parser_nested_simple_2(parser):
    assert 3.25 == parser('a+(b+c)/d-e+f',{'a':1, 'b':2, 'c':3, 'd':4, 'e':5, 'f':6, 'g':7, 'h':8})

def test_parser_nested(parser):
    assert 5.25 == parser('a+(b+c)/d-e+f-h/d*a+floor(sin(sqrt(g))*10)',{'a':1, 'b':2, 'c':3, 'd':4, 'e':5, 'f':6, 'g':7, 'h':8})


@pytest.mark.parametrize("expression, vars, expected", 
                         [  ('1+1', {}, 1+1), 
                            ('1-1', {}, 1-1),
                            ('1*1', {}, 1*1), 
                            ('1/1', {}, 1/1), 
                            ('1%1', {}, 0), 
                            ('1^1', {}, 1), 
                            ('1,1', {}, [1,1]), 
                            ('1||1',{}, '1.01.0'), 
                            ('1==1',{}, 1==1), 
                            ('1!=1',{}, 1!=1), 
                            ('1>1', {}, 1>1), 
                            ('1<1', {}, 1<1),
                            ('1<=1', {}, 1<=1),                                                         
                            ('1>=1', {}, 1>=1),                                                         
                            ('1 and 1', {}, 1 and 1), 
                            ('1 or 1', {}, 1 or 1)
                         ])
def test_parser_binary_operators(parser, expression,vars,expected):
    assert expected == parser(expression, vars)                        

@pytest.mark.parametrize("expression, vars, expected", 
                        [
                            ('sin(1)', {}, math.sin(1)),
                            ('cos(1)', {}, math.cos(1)),
                            ('tan(1)', {}, math.tan(1)),
                            ('asin(1)', {}, math.asin(1)),
                            ('acos(1)', {}, math.acos(1)),
                            ('atan(0.5)', {}, math.atan(0.5)),
                            ('sqrt(1)', {}, math.sqrt(1)),
                            ('log(1)', {}, math.log(1)),
                            ('abs(-1)', {}, abs(-1)),
                            ('ceil(0.3)', {}, math.ceil(0.3)),
                            ('floor(0.3)', {}, math.floor(0.3)),
                            ('round(1.6)', {}, round(1.6)),                                                                                                                                            
                            ('-(1)', {}, -(1)),
                            ('exp(1)', {}, math.exp(1)),
                            ('atan2(0.5,1)', {}, math.atan2(0.5,1)),                                                                                                                                           
                            ('fac(3)', {}, 3 * 2 * 1),
                            ('exp(1)-E', {}, 0),
                            ('PI/4 - atan(1)', {}, math.pi/4-math.atan(1))
                        ])
def test_parser_functions(parser, expression, vars, expected):
    assert expected == parser(expression, vars)


def test_parser_exceptions_div0(parser):
    with pytest.raises(ZeroDivisionError):
        parser('1/0', {})
    
def test_parser_exception_round_brackets(parser):
    with pytest.raises(ParsingError) as e:
        parser('(((1+1)+1)', {})

def test_parser_exception_not_a_function(parser):
    with pytest.raises(UndefinedVariableError) as e:
        parser('pippo', {})

def test_parser_exception_type_error(parser):
    with pytest.raises(TypeError):
        parser('1+sin(+sin())', {})

def test_parser_exception_unclosed_comment(parser):
    with pytest.raises(ValueError):
        parser('/* 1+1', {})

def test_parser_exception_bad_unicode(parser):
    with pytest.raises(ValueError):
        parser("concat('\u0H41', 'pippo')", {})

def test_parser_exception_bad_escape(parser):
    with pytest.raises(ParsingError):
        parser("concat('\z', 'pippo')", {})

@pytest.mark.bug
def test_parser_exception_var_not_found_bug(parser):
    with pytest.raises(UndefinedVariableError):
        assert 1 == parser('"__a__', {"__a__": 1})

@pytest.mark.bug        
def test_parser_escaped_chars(parser):
    with pytest.raises(ParsingError):
        parser("concat('\'', '\b', '\f', '\n', '\r', '\t', 'A')", {})        

def test_parser_this_is_a_comment_in_expression(parser):
    assert 2 == parser('1+ /* this is a comment */ 1', {})

@pytest.mark.bug
def test_parser_var_in_quotes_bug(parser):
    assert 1 == parser('"__a__"', {'"__a__"': 1})

def test_parser_unicode(parser):
    assert "AA" == parser("concat('\u0041', 'A')", {}) 







import pytest
from FormulaToolkit import DependencyResolver  

@pytest.fixture()
def dependency_resolver():
    dependency_resolver = DependencyResolver()
    yield dependency_resolver
    del dependency_resolver


def test_depencies_resolver_on_case_one(dependency_resolver):
    """This case is correct"""
    DEPENDENCIES_CASE = dict(
        a=('b', 'c'),
        d=('b' ), 
        c=('j', 'h', 'i', 'l'),
        e=('a', 'b', 'l', 'h'),
        f=('i', 'l' )
    )
    SOLUTION = [
        set(['i', 'h', 'b', 'l', 'j']),
        set(['c', 'd', 'f']),
        set(['a']),
        set(['e'])
    ]
    assert dependency_resolver.solve(DEPENDENCIES_CASE) == SOLUTION


def test_dependency_resolver_detect_loop(dependency_resolver):
    """LOOP"""
    DEPENDENCIES_CASE = dict(
        a=('b', 'c'),  # a depends on b and c 
        d=('b' ), 
        c=('a', 'h', 'i', 'l'), # c depends on a ( loop )
        e=('a', 'b', 'l', 'h'),
        f=('i', 'l' )
    )
    with pytest.raises(RuntimeError):
        dependency_resolver.solve(DEPENDENCIES_CASE)
        